"""Group2 Cookbook2."""
import argparse


__title__ = __doc__


def argument_parser():
    """As required by spicerack.cookbook."""
    parser = argparse.ArgumentParser('Argparse')
    parser.add_argument('-k', action='store_true')
    parser.add_argument('--argument')

    return parser


def run(args, spicerack):
    """As required by spicerack.cookbook."""
    print([args, spicerack.verbose, spicerack.dry_run])
    return 0



"""Prometheus module."""

import logging

from typing import Dict, List, Optional, Union

import requests

from spicerack.constants import ALL_DATACENTERS
from spicerack.exceptions import SpicerackError


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class PrometheusError(SpicerackError):
    """Custom exception class for errors of this module."""


class Prometheus:
    """Class to interact with the Prometheus server."""

    _prometheus_api = 'http://prometheus.svc.{site}.wmnet/ops/api/v1/query'

    def query(self, query: str, site: str, *, timeout: Optional[Union[float, int]] = 10) -> List[Dict]:
        """Preform a generic query.

        Arguments:
            query (str): a prometheus query
            site (str): The site to use for queries. Must be one of
                :py:const:`spicerack.constants.ALL_DATACENTERS`
            timeout (float, int, None, optional): How many seconds to wait for the prometheus to
                send data before giving up, as a float or int. Alternatively None to indicate an
                infinite timeout.

        Returns:
            list: returns an empty list if there are no results otherwise return a list of
            results of the form: {'metric': {}, 'value': [$timestamp, $value]}

        Raises:
            spicerack.prometheus.PrometheusError: on error

        """
        if site not in ALL_DATACENTERS:
            msg = 'site ({site}) must be one of spicerack.constants.ALL_DATACENTERS {dcs}'.format(
                site=site, dcs=ALL_DATACENTERS)
            raise PrometheusError(msg)

        url = self._prometheus_api.format(site=site)
        response = requests.get(url, params={'query': query}, timeout=timeout)
        if response.status_code != requests.codes['ok']:
            msg = 'Unable to get metric: HTTP {code}: {text}'.format(
                code=response.status_code, text=response.text)
            raise PrometheusError(msg)

        result = response.json()

        if result.get('status', 'error') == 'error':
            msg = 'Unable to get metric: {error}'.format(error=result.get('error', 'unknown'))
            raise PrometheusError(msg)

        return result['data']['result']



"""Dnsdisc module tests."""
from collections import namedtuple
from unittest import mock

import pytest

from dns.exception import DNSException

from spicerack.dnsdisc import Discovery, DiscoveryCheckError, DiscoveryError
from spicerack.tests import require_caplog


MockedRecord = namedtuple('Record', ['address'])


def mock_obj(datacenter, svc, pooled):
    """Creates a mock conftool object."""
    obj = mock.MagicMock()
    obj.tags = {'dnsdisc': svc}
    obj.name = datacenter
    obj.pooled = pooled
    return obj


class MockedQuery:
    """Class to mock a return object from a call to dns.resolver.query()."""

    def __init__(self, record):
        """Initialize it with the query."""
        if record == 'fail.svc.eqiad.wmnet':
            self.address = '10.1.1.1'
            self.ttl = 600
        elif record.startswith('raise'):
            raise DNSException
        else:
            self.address = '10.0.0.1'
            self.ttl = 10

        self.record = MockedRecord(self.address)

    def __getitem__(self, _):
        """Allow indexing access."""
        return self.record


class TestDiscovery:
    """Discovery class tests."""

    @mock.patch('spicerack.dnsdisc.resolver')
    def setup_method(self, _, mocked_resolver):
        """Initialize the test environment for Discovery."""
        # pylint: disable=attribute-defined-outside-init
        self.records = ['record1', 'record2']
        self.conftool_records = '({records})'.format(records='|'.join(self.records))
        self.nameservers = ['authdns1', 'authdns2']

        self.mocked_confctl = mock.MagicMock()
        self.mocked_remote = mock.MagicMock()
        self.mocked_remote.query.return_value.hosts = self.nameservers

        self.mocked_resolver = mocked_resolver
        self.mocked_resolver.Resolver.return_value.query = MockedQuery

        self.discovery = Discovery(self.mocked_confctl, self.mocked_remote, self.records, dry_run=False)
        self.discovery_dry_run = Discovery(self.mocked_confctl, self.mocked_remote, self.records)

    def test_init(self):
        """Creating a Discovery instance should initialize the resolvers based on a cumin query."""
        self.mocked_remote.query.assert_has_calls([mock.call('A:dns-auth')] * 2)
        self.mocked_resolver.Resolver.assert_called_with()
        assert self.mocked_resolver.Resolver.call_count == 4
        self.mocked_resolver.query.assert_has_calls(
            [mock.call(self.nameservers[0]), mock.call(self.nameservers[1])], any_order=True)
        assert self.mocked_resolver.query.call_count == 4

    def test_init_ko(self):
        """Creating a Discovery instance should raise DiscoveryError if unable to initialize the resolvers."""
        self.mocked_resolver.Resolver.return_value.query.side_effect = DNSException
        with pytest.raises(DiscoveryError, match='Unable to resolve authdns1'):
            Discovery(self.mocked_confctl, self.mocked_remote, self.records, dry_run=False)

    def test_resolve_address(self):
        """Calling resolve_address() sould raise DiscoveryError if unable to resolve the address."""
        with pytest.raises(DiscoveryError, match='Unable to resolve raise.svc.eqiad.wmnet'):
            self.discovery.resolve_address('raise.svc.eqiad.wmnet')

    def test_update_ttl(self):
        """Calling update_ttl() should update the TTL of the conftool objects."""
        self.discovery.update_ttl(10)
        self.mocked_confctl.assert_has_calls([mock.call.set_and_verify('ttl', 10, dnsdisc=self.conftool_records)])

    @mock.patch('spicerack.decorators.time.sleep')
    def test_update_ttl_dry_run(self, mocked_sleep):
        """Calling update_ttl() in DRY-RUN mode should not raise when verifying the TTL."""
        self.discovery_dry_run.update_ttl(20)
        assert not mocked_sleep.called

    @mock.patch('spicerack.decorators.time.sleep')
    def test_update_ttl_ko(self, mocked_sleep):
        """Calling update_ttl() should raise DiscoveryCheckError if unable to verify the value."""
        with pytest.raises(DiscoveryCheckError, match="Expected TTL '20', got '10'"):
            self.discovery.update_ttl(20)

        assert mocked_sleep.called

    def test_check_ttl_ok(self):
        """Calling check_ttl() should verify that the correct TTL is returned by the authoritative nameservers."""
        self.discovery.check_ttl(10)

    @mock.patch('spicerack.decorators.time.sleep')
    def test_check_ttl_ko(self, mocked_sleep):
        """Calling check_ttl() should raise DiscoveryCheckError if the check fails."""
        with pytest.raises(DiscoveryCheckError, match="Expected TTL '20', got '10'"):
            self.discovery.check_ttl(20)

        assert mocked_sleep.called

    def test_check_record_ok(self):
        """Calling check_record() should verify that a record has a certain value on the nameservers."""
        self.discovery.check_record(self.records[0], 'ok.svc.eqiad.wmnet')

    @mock.patch('spicerack.decorators.time.sleep')
    def test_check_record_ko(self, mocked_sleep):
        """Calling check_record() should raise DiscoveryError if unable to check the records."""
        with pytest.raises(DiscoveryError, match='Failed to check record {record}'.format(record=self.records[0])):
            self.discovery.check_record(self.records[0], 'fail.svc.eqiad.wmnet')

        assert mocked_sleep.called

    def test_resolve(self):
        """Calling resolve() should raise a DiscoveryError if unable to resolve the address."""
        with pytest.raises(DiscoveryError, match='Unable to resolve raise.discovery.wmnet from authdns'):
            for _ in self.discovery.resolve(name='raise'):
                pass

    @pytest.mark.parametrize('func, value', (('pool', True), ('depool', False)))
    def test_pool(self, func, value):
        """Calling pool() should update the pooled value of the conftool objects to True."""
        getattr(self.discovery, func)('eqiad')
        self.mocked_confctl.set_and_verify.assert_called_once_with(
            'pooled', value, dnsdisc=self.conftool_records, name='eqiad')

    def test_active_datacenters(self):
        """The list of active datacenter is correctly composed."""
        self.mocked_confctl.get.return_value = [
            mock_obj('dcA', 'svcA', True),
            mock_obj('dcB', 'svcA', False),
            mock_obj('dcA', 'svcB', True),
            mock_obj('dcB', 'svcB', True)
        ]
        expected = {'svcA': ['dcA'], 'svcB': ['dcA', 'dcB']}
        assert expected == self.discovery.active_datacenters
        self.mocked_confctl.get.assert_called_once_with(dnsdisc=self.conftool_records)

    def test_check_if_depoolable_ok(self):
        """No exception is raised if the active datacenters are ok."""
        self.mocked_confctl.get.return_value = [
            mock_obj('dcA', 'svcA', False),
            mock_obj('dcB', 'svcA', True),
            mock_obj('dcA', 'svcB', True),
            mock_obj('dcB', 'svcB', True)
        ]
        self.discovery.check_if_depoolable('dcA')
        # now let's assume a service is competely down. We should still get
        # a green light.
        self.mocked_confctl.get.return_value = [
            mock_obj('dcA', 'svcA', False),
            mock_obj('dcB', 'svcA', False),
            mock_obj('dcA', 'svcB', True),
            mock_obj('dcB', 'svcB', True)
        ]
        self.discovery.check_if_depoolable('dcA')

    def test_check_if_depoolable_ko(self):
        """A DiscoveryError is raised when a service would be taken out of commission."""
        self.mocked_confctl.get.return_value = [
            mock_obj('dcA', 'svcA', False),
            mock_obj('dcB', 'svcA', True),
            mock_obj('dcA', 'svcB', False),
            mock_obj('dcB', 'svcB', False)
        ]
        with pytest.raises(DiscoveryError,
                           match='Services svcA cannot be depooled as they are only active in dcB'):
            self.discovery.check_if_depoolable('dcB')
        self.mocked_confctl.get.return_value = [
            mock_obj('dcA', 'svcA', False),
            mock_obj('dcB', 'svcA', True),
            mock_obj('dcA', 'svcB', False),
            mock_obj('dcB', 'svcB', True)
        ]
        with pytest.raises(DiscoveryError,
                           match='Services svcA, svcB cannot be depooled as they are only active in dcB'):
            self.discovery.check_if_depoolable('dcB')

    @require_caplog
    def test_check_if_depoolable_ko_dry_run(self, caplog):
        """Doesn't raise exception when a service would be taken out of commission but in dry-run mode."""
        self.mocked_confctl.get.return_value = [
            mock_obj('dcA', 'svcA', False),
            mock_obj('dcB', 'svcA', True),
            mock_obj('dcA', 'svcB', False),
            mock_obj('dcB', 'svcB', False)
        ]
        self.discovery_dry_run.check_if_depoolable('dcB')
        assert 'cannot be depooled as they are only active in' in caplog.text



"""Interactive module."""
import getpass
import logging
import os
import sys

from spicerack.exceptions import SpicerackError


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
MIN_SECRET_SIZE = 6


def ask_confirmation(message: str) -> None:
    """Ask the use for confirmation in interactive mode.

    Arguments:
        message (str): the message to be printed before asking for confirmation.

    Raises:
        SpicerackError: on too many invalid answers or if not in a TTY.

    """
    if not sys.stdout.isatty():
        raise SpicerackError('Not in a TTY, unable to ask for confirmation')

    print(message)
    print('Type "done" to proceed')

    for _ in range(3):
        resp = input('> ')
        if resp == 'done':
            break

        print('Invalid response, please type "done" to proceed. After 3 wrong answers the task will be aborted.')
    else:
        raise SpicerackError('Too many invalid confirmation answers')


def get_username() -> str:
    """Detect and return the name of the effective running user even if run as root.

    Returns:
        str: the name of the effective running user or ``-`` if unable to detect it.

    """
    user = os.getenv('USER')
    sudo_user = os.getenv('SUDO_USER')

    if sudo_user is not None and sudo_user != 'root':
        return sudo_user

    if user is not None:
        return user

    return '-'


def ensure_shell_is_durable() -> None:
    """Ensure it is running either in non-interactive mode or in a screen/tmux session, raise otherwise.

    Raises:
        spicerack.exceptions.SpicerackError: if in a non-durable shell session.

    """
    # STY is for screen, TMUX is for tmux. Not using `getenv('NAME') is not None` to check they are not empty.
    # TODO: verify if the check on TERM is redundant.
    if (sys.stdout.isatty() and not os.getenv('STY', '') and not os.getenv('TMUX', '')
            and 'screen' not in os.getenv('TERM', '')):
        raise SpicerackError('Must be run in non-interactive mode or inside a screen or tmux.')


def get_management_password() -> str:
    """Get the management password either from the environment or asking for it.

    Returns:
        str: the password.

    Raises:
        spicerack.exceptions.SpicerackError: if the password is empty.

    """
    password = os.getenv('MGMT_PASSWORD')

    if password is None:
        logger.debug('MGMT_PASSWORD environment variable not found')
        # Ask for a password, raise exception if not a tty
        password = get_secret('Management Password')
    else:
        logger.info('Using Management Password from the MGMT_PASSWORD environment variable')

    if not password:
        raise SpicerackError('Empty Management Password')

    return password


def get_secret(title: str, *, confirm: bool = False) -> str:
    """Ask the user for a secret e.g. password.

    Arguments:
        title (str): The message to show the user.
        confirm (bool, optional): If :py:data:`True` ask the user to confirm the password.

    Returns:
        str: the secret.

    """
    new_secret = getpass.getpass(prompt='{}: '.format(title))

    while len(new_secret) < MIN_SECRET_SIZE:
        new_secret = getpass.getpass(
            prompt='Secret must be at least {} characters. try again: '.format(MIN_SECRET_SIZE))

    if confirm and new_secret != getpass.getpass(prompt='Again, just to be sure: '):
        raise SpicerackError('{}: Passwords did not match'.format(title))

    return new_secret



"""Redis cluster module."""
import logging
import os

from collections import defaultdict
from typing import Any, Tuple, Union

from redis import StrictRedis

from spicerack.config import load_yaml_config
from spicerack.exceptions import SpicerackError


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class RedisClusterError(SpicerackError):
    """Custom exception class for errors in the RedisCluster class."""


class RedisCluster:
    """Class to manage a Redis Cluster."""

    def __init__(self, cluster: str, config_dir: str, *, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            cluster (str): the name of the cluster to connect to.
            config_dir (str): path to the directory containing the configuration files for the Redis clusters.
            dry_run (bool, optional): whether this is a DRY-RUN.

        """
        self._dry_run = dry_run
        self._shards = defaultdict(dict)  # type: ignore
        config = load_yaml_config(os.path.join(config_dir, cluster + '.yaml'))

        for datacenter, shards in config['shards'].items():
            for shard, data in shards.items():
                self._shards[datacenter][shard] = RedisInstance(
                    host=data['host'], port=data['port'], password=config['password'], decode_responses=True)

    def start_replica(self, datacenter: str, master_datacenter: str) -> None:
        """Start the cluster replica in a datacenter from a master datacenter.

        Arguments:
            datacenter (str): the datacenter on which to start the replica.
            master_datacenter (str): the datacenter from which to replicate.

        Raises:
            RedisClusterError: on error and invalid parameters.

        """
        if master_datacenter == datacenter:
            raise RedisClusterError(
                'Master datacenter must be different from the current datacenter, got {dc}'.format(dc=datacenter))

        for shard, instance in sorted(self._shards[datacenter].items()):
            self._start_instance_replica(instance, self._shards[master_datacenter][shard])

    def stop_replica(self, datacenter: str) -> None:
        """Stop the cluster replica in a datacenter.

        Arguments:
            datacenter (str): the datacenter on which to stop the replica.

        Raises:
            RedisClusterError: on error.

        """
        for instance in self._shards[datacenter].values():
            self._stop_instance_replica(instance)

    def _start_instance_replica(self, instance: 'RedisInstance', master: 'RedisInstance') -> None:
        """Start the replica in a specific instance from a master instance.

        Arguments:
            instance (spicerack.redis_cluster.RedisInstance): the instance where to start the replica.
            master (spicerack.redis_cluster.RedisInstance): the master instance to replicate from.

        Raises:
            RedisClusterError: if unable to verify the replica has started.

        """
        if instance.master_info == master.info:
            logger.debug('Replica already configured on %s', instance)
            return

        if self._dry_run:
            logger.debug('Skip starting replica on %s in dry-run mode', instance)
        else:
            logger.debug('Starting replica %s => %s', master, instance)
            instance.start_replica(master)

        if not self._dry_run and instance.master_info != master.info:
            raise RedisClusterError('Replica on {instance} is not correctly configured: {parent}'.format(
                instance=instance, parent=instance.master_info))

    def _stop_instance_replica(self, instance: 'RedisInstance') -> None:
        """Stop the replica in a specific instance.

        Arguments:
            instance (spicerack.redis_cluster.RedisInstance): the instance where to stop the replica.

        Raises:
            RedisClusterError: on error.

        """
        if instance.is_master:
            logger.debug('Instance %s is already master, doing nothing', instance)
            return

        if self._dry_run:
            logger.debug('Skip stopping replica on %s in dry-run mode', instance)
        else:
            logger.debug('Stopping replica on %s', instance)
            instance.stop_replica()

        if not self._dry_run and not instance.is_master:
            raise RedisClusterError('Instance {instance} is still a slave of {parent}, aborting'.format(
                instance=instance, parent=instance.master_info))


class RedisInstance:
    """Class to manage a Redis instance, a simple wrapper around `redis.StrictRedis`."""

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the instance.

        Arguments:
            **kwargs (mixed): arbitrary keyword arguments, to be passed to the `redis.StrictRedis` constructor.

        """
        self.host = kwargs.get('host')
        self.port = kwargs.get('port')
        self._client = StrictRedis(**kwargs)

    @property
    def is_master(self) -> bool:
        """Getter to check if the current instance is a master.

        Returns:
            bool: :py:data:`True` if the instance is a master, :py:data:`False` otherwise.

        """
        return self._client.info('replication')['role'] == 'master'

    @property
    def master_info(self) -> Union[Tuple[None, None], Tuple[str, int]]:
        """Getter to know the master of this instance.

        Returns:
            tuple: a 2-element tuple with (host/IP, port) of the master instance. If there is no master configured
            (None, None) is returned.

        """
        data = self._client.info('replication')
        try:
            return data['master_host'], data['master_port']
        except (KeyError, TypeError):
            return (None, None)

    @property
    def info(self) -> Tuple[str, int]:
        """Getter to know the detail of this instance.

        Returns:
            tuple: a 2-element tuple with (host/IP, port) of the instance.

        """
        return self.host, self.port  # type: ignore

    def stop_replica(self) -> None:
        """Stop the replica on the instance."""
        self._client.slaveof()

    def start_replica(self, master: 'RedisInstance') -> None:
        """Start the replica from the given master instance.

        Arguments:
            master (spicerack.redis_cluster.RedisInstance): the master instance.

        """
        self._client.slaveof(master.host, master.port)

    def __str__(self) -> str:
        """String representation of the instance.

        Returns:
            str: the host or IP and port of the instance.

        """
        return '{0.host}:{0.port}'.format(self)



"""Actions module."""
import logging

from collections import OrderedDict
from typing import Hashable, List


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Actions:
    """Class to keep track and log a set of actions performed and their result with a nice string representation."""

    def __init__(self, name: Hashable):
        """Initialize the instance.

        When converted to string returns a nicely formatted representation of the instance and all its actions.

        It exposes the following properties:

        - ``name``: the name passed to the instance at instantiation time.

        - ``has_warnings``: a :py:class:`bool` that is :py:data:`True` when at least one warning action was registered,
          :py:data:`False` otherwise.

        - ``has_failures``: a :py:class:`bool` that is :py:data:`True` when at least one failed action was registered,
          :py:data:`False` otherwise.

        Arguments:
            name (typing.Hashable): the name of the set of actions to be registered.

        """
        self.name = name
        self.actions = []  # type: List[str]
        self.has_warnings = False
        self.has_failures = False

    def __str__(self) -> str:
        """Custom string representation of the actions performed.

        Returns:
            str: the string representation.

        """
        actions = '\n'.join('  - {action}'.format(action=action) for action in self.actions)
        return '{name} (**{status}**)\n{actions}'.format(name=self.name, status=self.status, actions=actions)

    @property
    def status(self) -> str:
        """Return the current status of the actions based on the worst result recorded.

        Returns:
            str: the short string representation of the status, one of: ``PASS``, ``WARN``, ``FAIL``.

        """
        if self.has_failures:
            return 'FAIL'
        if self.has_warnings:
            return 'WARN'

        return 'PASS'

    def success(self, message: str) -> None:
        """Register a successful action.

        Arguments:
            message (str): the action description.

        """
        self._action(logging.INFO, message)

    def failure(self, message: str) -> None:
        """Register a failed action.

        Arguments:
            message (str): the action description.

        """
        self._action(logging.ERROR, message)
        self.has_failures = True

    def warning(self, message: str) -> None:
        """Register an action that require some attention.

        Arguments:
            message (str): the action description.

        """
        self._action(logging.WARNING, message)
        self.has_warnings = True

    def _action(self, level: int, message: str) -> None:
        """Register a generic action.

        Arguments:
            level (int): a logging level integer to register the action for.
            message (str): the action description.

        """
        logger.log(level, message)
        self.actions.append(message)


class ActionsDict(OrderedDict):
    """OrderedDict with defaultdict capabilities for the :py:class:`spicerack.actions.Action` class.

    Inherits from :py:class:`collections.OrderedDict` and automatically instantiate and returns a new instance of the
    :py:class:`spicerack.actions.Actions` class for every missing key like a :py:class:`collections.defaultdict`.

    When accessing a missing key, the key itself is passed to the new :py:class:`spicerack.actions.Actions` instance
    as ``name``.

    When converted to string returns a nicely formatted representation of the instance and all its items.

    Todo:
        Once the support of Python 3.5 is dropped this can be converted into a simple factory with a single ``new()``
        method that returns an instance of ``defaultdict(Actions)``.

    """

    def __missing__(self, key: Hashable) -> Actions:
        """Instantiate a new Actions instance for the missing key like a defaultdict.

        Parameters as required by Python's data model, see :py:method:`object.__missing__`.

        """
        self[key] = Actions(key)
        return self[key]

    def __str__(self) -> str:
        """Custom string representation of the instance.

        Returns:
            str: the string representation.

        """
        return '\n'.join('- {actions}\n'.format(actions=value) for key, value in self.items())



"""Constants module."""

ALL_DATACENTERS = ('eqiad', 'codfw', 'esams', 'ulsfo', 'eqsin')
"""tuple: list of all datacenters."""

CORE_DATACENTERS = ('eqiad', 'codfw')
"""tuple: list of core datacenters."""

INTERNAL_TLD = 'wmnet'
"""str: the name of the internal top level domain."""

MANAGEMENT_SUBDOMAIN = 'mgmt'
"""str: the name of the subdomain for the management interfaces."""

PUPPET_CA_PATH = '/etc/ssl/certs/Puppet_Internal_CA.pem'
"""str: the path to the Puppet Signing CA cert"""



"""Cookbook module."""
import argparse
import importlib
import logging
import os
import shlex
import sys

from abc import abstractmethod
from typing import Dict, List, Optional, Tuple, Type, Union

from spicerack import log, Spicerack
from spicerack.config import load_yaml_config
from spicerack.exceptions import SpicerackError


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
COOKBOOK_NO_PARSER_WITH_ARGS_RETCODE = 95
"""int: reserved exit code: the cookbook doesn't have an argument_parser() function but was called with arguments."""
COOKBOOK_PARSE_ARGS_FAIL_RETCODE = 96
"""int: reserved exit code: the cookbook fail to parse arguments."""
COOKBOOK_INTERRUPTED_RETCODE = 97
"""int: reserved exit code: the execution was interrupted."""
COOKBOOK_NOT_FOUND_RETCODE = 98
"""int: reserved exit code: no cookbook is found for the selection."""
COOKBOOK_EXCEPTION_RETCODE = 99
"""int: reserved exit code: the cookbook raised an exception while executing."""
COOKBOOKS_MENU_HELP_MESSAGE = """Cookbooks interactive menu help

Available cookbooks and cookbook groups are shown in the menu with the format:
  [STATUS] NAME: DESCRIPTION
Additional control commands are also shown.
To select an item just input its name and press Enter.

Group of cookbooks:
  They have a status that represent the number of the executed cookbooks over
  the total number of cookbooks in that group and its child groups
  (i.e. [2/11]) or the status 'DONE' in case all cookbooks in that group were
  executed during the current session.
  When selected the child cookbooks group is shown.

Single cookbooks:
  Their status has one of the following values:
  {statuses}
  When selected the cookbook is executed and then the current menu is shown
  again after its execution, with the status updated based on the result of the
  execution.

Control commands:
  b: shown when inside a child group of cookbooks to go back one level to the
     parent menu.
  q: shown when at the top level of the current session menu to exit the
     program.
  h: always shown, print this help message.

  Note: 'q' and 'b' are mutually exclusive, only one of them is shown.

CLI arguments:
  It's possible to pass CLI parameters to cookbooks and group of cookbooks
  when selecting them (e.g. cookbook_name -a param value1 value2).
  Passing arguments to cookbook groups propagate them also to their cookbooks
  and child cookbook groups.
  Passing arguments override any other argument that might have been passed
  to the cookbook executable or to any of the parent groups when selected.

Interrupting execution:
  Pressing Ctrl+c/d while executing a cookbook interrupts it and show the
  current menu, marking the cookbook status as ERROR.
  Pressing Ctrl+c/d while in a menu is equivalent to select 'b' or 'q'.
"""
"""str: the generic CookbooksMenu help message, unformatted."""


class CookbookError(SpicerackError):
    """Custom exception class for errors of this module."""


class CookbooksModuleInterface:
    """Module interface to be used as type hint for the imported cookbooks."""

    __title__ = ''
    """str: the cookbook static title."""

    @staticmethod
    def argument_parser() -> argparse.ArgumentParser:
        """Optional module function to define if the cookbook should accept command line arguments."""

    @staticmethod
    def run(args: Optional[argparse.Namespace], spicerack: Spicerack) -> Optional[int]:
        """Mandatory module function that every cookbook must define with the cookbook's body."""


class Cookbooks:
    """Collect and represent available cookbooks."""

    cookbooks_module_prefix = 'cookbooks'

    def __init__(self, base_dir: str, args: List[str], spicerack: Spicerack, path_filter: Optional[str] = None) -> None:
        """Initialize the Cookbook class and collect CookbooksMenu and Cookbook items.

        Arguments:
            base_dir (str): the base directory from where to start looking for cookbooks.
            args (list): the list of arguments to pass to the collected items.
            spicerack (spicerack.Spicerack): the initialized instance of the library.
            path_filter (str, optional): an optional relative module path to filter for. If set, only cookbooks that
                are part of this subtree will be collected.

        """
        self.base_dir = os.path.join(base_dir, self.cookbooks_module_prefix)
        self.args = args
        self.spicerack = spicerack
        if path_filter is None:
            self.path_filter = path_filter
        else:
            self.path_filter = '.'.join((self.cookbooks_module_prefix, path_filter))

        self.menu = CookbooksMenu(self.cookbooks_module_prefix, self.args, self.spicerack)
        self._collect()

    def get_item(self, path: str) -> Optional[Union['CookbooksMenu', 'Cookbook']]:
        """Retrieve the item for a given path.

        Arguments:
            path (str): the path of the item to look for.

        Returns:
            None: when no item is found.
            spicerack.cookbook.CookbooksMenu: when the item found is a menu of cookbooks.
            spicerack.cookbook.Cookbook: when the item found is a single cookbook.

        """
        item = None
        progressive_path = []
        for i, subpath in enumerate(path.split('.')):
            progressive_path.append(subpath)
            if i == 0:
                item = self.menu
            elif item is not None and subpath in item.items.keys():
                item = item.items[subpath]
            else:
                item = None

        return item

    def _create_menu_for_path(self, path: str) -> 'CookbooksMenu':
        """Create the menu for a given path, including intermediate levels, if missing. Return the existing one if any.

        Arguments:
            path (str): the path of the item to look for.

        Returns:
            spicerack.cookbook.CookbooksMenu: the existing or created menu.

        """
        item = self.menu
        if path == self.cookbooks_module_prefix:
            return item

        progressive_path = [self.cookbooks_module_prefix]
        for subpath in path.split('.')[1:]:
            progressive_path.append(subpath)
            if item is not None and subpath in item.items.keys():
                item = item.items[subpath]
            else:
                module_name = '.'.join(progressive_path)
                submenu = CookbooksMenu(module_name, self.args, self.spicerack)
                item.append(submenu, add_parent=self._submenu_add_parent(module_name, path))
                item = submenu

        return item

    def _submenu_add_parent(self, module_name: str, path: str) -> bool:
        """Determine if the submenu item to be appended should have a link to the parent menu or not.

        When collecting the cookbooks and creating the CookbooksMenu instances, the relation to the parent menu should
        be skipped for those intermediate menus created for coherence but that should not be accessible by the user,
        like when using a path_filter.

        Arguments:
            module_name (str): the module name of the submenu.
            path (str): of the item to be add.

        Returns:
            bool: True if the link to the parent menu should be set, False otherwise.

        """
        if module_name != path:
            return False

        if self.path_filter is None:
            return True

        return path.startswith(self.path_filter) and len(path) > len(self.path_filter)

    @staticmethod
    def _filter_dirnames_and_filenames(dirnames: List[str], filenames: List[str]) -> Tuple[List, List]:
        """Filter the dirnames and filenames in place (required by os.walk()) to select only Python modules.

        Arguments:
            dirnames (list): the list of sub-directories, as returned by os.walk().
            filenames (list): the list of filenames in the current directory, as returned by os.walk().

        Returns:
            tuple: (list, list) with the modified dirnames and filenames.

        """
        try:
            dirnames.remove('__pycache__')
        except ValueError:
            pass

        for filename in filenames.copy():
            if filename == '__init__.py' or not filename.endswith('.py'):
                filenames.remove(filename)

        return dirnames, filenames

    def _collect(self) -> None:
        """Collect available cookbooks starting from a base path."""
        for dirpath, dirnames, filenames in os.walk(self.base_dir):
            dirnames, filenames = Cookbooks._filter_dirnames_and_filenames(dirnames, filenames)
            if not filenames and not dirnames:
                continue

            # Sort the directories and files to be recursed in-place, required by os.walk().
            dirnames.sort()
            filenames.sort()

            relpath = os.path.relpath(dirpath, start=self.base_dir)
            prefix = self.cookbooks_module_prefix
            if relpath != '.':
                prefix = '.'.join((self.cookbooks_module_prefix, relpath.replace('/', '.')))

            if (self.path_filter is not None and not prefix.startswith(self.path_filter)
                    and sum('.'.join((prefix, filename)).startswith(self.path_filter) for filename in filenames) == 0):
                continue

            path = prefix.rstrip('.')
            try:
                menu = self._create_menu_for_path(path)
            except CookbookError as e:
                logger.error(e)
                continue

            for filename in filenames:
                self._collect_filename(filename, prefix, menu)

    def _collect_filename(self, filename: str, prefix: str, menu: 'CookbooksMenu') -> None:
        """Iterate the filenames in the current directory as reported by os.walk() and add them to the tree.

        Arguments:
            filename (str): the filename to collect.
            prefix (str): the Python module prefix to use to load the given filename.
            menu (spicerack.cookbook.CookbooksMenu): the menu to append the collected cookbook to.

        """
        cookbook_module_name = '.'.join((prefix, os.path.splitext(filename)[0]))
        if self.path_filter is not None and not cookbook_module_name.startswith(self.path_filter):
            return

        try:
            cookbook = Cookbook(cookbook_module_name, self.args, self.spicerack)
        except CookbookError as e:
            logger.error(e)
            return

        menu.append(cookbook)


class BaseCookbooksItem:
    """Base class for any item collected by the Cookbooks class."""

    fallback_title = '-'

    def __init__(self, module_name: str, args: List[str], spicerack: Spicerack) -> None:
        """Base cookbooks's item constructor.

        Arguments:
            module_name (str): the Python module to load.
            args (list): the command line arguments to pass to the item.
            spicerack (spicerack.Spicerack): the initialized instance of the library.

        """
        if '.' in module_name:
            self.name = module_name.rsplit('.', 1)[1]
            self.path = module_name.split('.', 1)[1]
        else:
            self.name = module_name
            self.path = module_name

        self.module = import_module(module_name)
        self.args = args
        self.spicerack = spicerack
        self.title = self._get_title()

    @abstractmethod
    def run(self) -> int:
        """Excecute the item."""

    @property
    def verbose_title(self) -> str:
        """Getter for the verbose_title property, uses the module name if there is no title.

        Returns:
            str: the verbose title of the item.

        """
        if self.title != self.fallback_title:
            return self.title

        return self.name

    def _get_title(self) -> str:
        """Calculate the title of the instance item.

        Returns:
            str: the title of the item.

        """
        try:
            title = self.module.__title__
        except AttributeError as e:
            logger.debug('Unable to detect title for module %s: %s', self.path, e)
            title = self.fallback_title

        return title

    @staticmethod
    def _get_line_prefix(level: int, cont_levels: List[bool], is_final: bool) -> str:
        """Return the line prefix for the given level in the tree.

        Arguments:
            level (int): how many levels the item is nested in the tree.
            cont_levels (list): a list of size levels with booleans that indicate for each level if the
                continuation prefix (True) or an empty prefix (False) should be used.
            is_final (bool): whether the line is the final in its own group.

        Returns:
            str: the line prefix to use.

        """
        empty_sep = '    '
        cont_sep = '|   '

        if is_final:
            base_sep = '`-- '
        else:
            base_sep = '|-- '

        if level == 0:
            return base_sep

        levels = []
        for cont_level in cont_levels:
            if cont_level:
                levels.append(cont_sep)
            else:
                levels.append(empty_sep)

        return ''.join(levels) + base_sep


class Cookbook(BaseCookbooksItem):
    """Cookbook class."""

    fallback_title = 'UNKNOWN (unable to detect title)'
    statuses = ('NOTRUN', 'PASS', 'FAIL', 'ERROR')  # Status labels
    not_run, success, failed, error = statuses  # Valid statuses variables

    def __init__(self, module_name: str, args: List[str], spicerack: Spicerack) -> None:
        """Override parent constructor to add menu-specific initialization.

        :Parameters:
            according to spicerack.cookbook.BaseCookbooksItem.

        """
        super().__init__(module_name, args, spicerack)
        self.status = Cookbook.not_run

    def run(self) -> int:
        """Run the cookbook, calling both its `argument_parser` and `run` functions.

        Returns:
            int: the return code to use for this cookbook, it should be zero on success, a positive integer smaller than
            128 and not in the range 90-99 (reserved exit codes) in case of failure.

        """
        ret, args = self._parse_args()
        if ret != -1:
            return ret

        return self._run(args)

    def _run(self, args: Optional[argparse.Namespace]) -> int:
        """Run the cookbook's `run()` function.

        Arguments:
            args (argparse.Namespace, None): the parsed arguments or None if the cookbook doesn't define a
                `argument_parser()` function.

        Returns:
            int: the return code to use for this cookbook, it should be zero on success, a positive integer smaller than
            128 and not in the range 90-99 (reserved exit codes) in case of failure.

        """
        log.log_task_start('Cookbook ' + self.path)
        message = 'raised while executing cookbook'

        try:
            ret = self.module.run(args, self.spicerack)
            if ret is None:
                ret = 0
        except KeyboardInterrupt:
            logger.error('Ctrl+c pressed')
            self.status = Cookbook.error
            ret = COOKBOOK_INTERRUPTED_RETCODE
        except SystemExit as e:
            if isinstance(e.code, int):
                ret = e.code
                if e.code == 0:
                    logger.info('SystemExit(0) %s %s, assuming success:', message, self.path)
                    self.status = Cookbook.success
                else:
                    logger.exception('SystemExit(%d) %s %s:', e.code, message, self.path)
                    self.status = Cookbook.error
            else:
                logger.exception("SystemExit('%s') %s %s:", e.code, message, self.path)
                self.status = Cookbook.error
                ret = COOKBOOK_EXCEPTION_RETCODE
        except BaseException:  # pylint: disable=broad-except
            logger.exception('Exception %s %s:', message, self.path)
            self.status = Cookbook.failed
            ret = COOKBOOK_EXCEPTION_RETCODE
        else:
            self.status = Cookbook.success if ret == 0 else Cookbook.failed

        log.log_task_end(self.status, 'Cookbook {name} (exit_code={ret})'.format(
            name=self.path, ret=ret))

        return ret

    def _parse_args(self) -> Tuple[int, Optional[argparse.Namespace]]:
        """Get the argument parser from the cookbook, if it exists, and parse the arguments.

        Returns:
            tuple: (int, argparse.Namespace) with the return code to use and the parsed arguments. If the return code is
            different from -1 it means that the cookbook should not be executed either because the help message was
            requested or the parse of the arguments failed or arguments were passed but the cookbook doesn't define
            a argument_parser() function.

        """
        ret = -1
        args = None
        message = 'raised while parsing arguments for cookbook'

        if not hasattr(self.module, 'argument_parser'):
            if self.args:
                ret = COOKBOOK_NO_PARSER_WITH_ARGS_RETCODE

            return ret, args

        try:
            args = self.module.argument_parser().parse_args(self.args)
        except SystemExit as e:
            if isinstance(e.code, int):
                ret = e.code
            else:
                logger.exception("SystemExit('%s') %s %s:", e.code, message, self.path)
                ret = COOKBOOK_PARSE_ARGS_FAIL_RETCODE
        except BaseException:  # pylint: disable=broad-except
            logger.exception('Exception %s %s:', message, self.path)
            ret = COOKBOOK_PARSE_ARGS_FAIL_RETCODE

        return ret, args


class CookbooksMenu(BaseCookbooksItem):
    """Cookbooks Menu class."""

    back_answer = 'b'
    """str: interactive menu answer to go back to the parent menu."""
    help_answer = 'h'
    """str: interactive menu answer to print the generic CookbooksMenu help message."""
    quit_answer = 'q'
    """str: answer to quit the interactive menu."""
    help_message = COOKBOOKS_MENU_HELP_MESSAGE.format(statuses=Cookbook.statuses)
    """str: the generic CookbooksMenu help message."""

    def __init__(self, module_name: str, args: List[str], spicerack: Spicerack) -> None:
        """Override parent constructor to add menu-specific initialization.

        :Parameters:
            according to spicerack.cookbook.BaseCookbooksItem.

        """
        super().__init__(module_name, args, spicerack)
        self.parent = None  # type: Optional[CookbooksMenu]
        self.items = {}  # type: ignore

    @property
    def status(self) -> str:
        """Getter for the menu status, returns a string representation of the status of its tasks.

        Returns:
            str: the menu status message.

        """
        completed, total = self.calculate_status()
        if completed == total:
            message = 'DONE'
        else:
            message = '{completed}/{total}'.format(completed=completed, total=total)

        return message

    def append(self, item: Union['CookbooksMenu', Cookbook], add_parent: bool = True) -> None:
        """Append an item to this menu.

        Arguments:
            item (spicerack.cookbook.Cookbook, spicerack.cookbook.CookbooksMenu): the item to append.
            add_parent (bool, optional): wheter to set the parent of the new item to the current instance.

        """
        if add_parent and isinstance(item, CookbooksMenu):
            item.parent = self

        self.items[item.name] = item

    def run(self) -> int:
        """Execute the menu in an interactive way (infinite loop).

        Returns:
            int: being an interactive menu it always returns 0.

        """
        try:
            while True:
                self.run_once()
        except StopIteration:
            pass

        return 0

    def show(self) -> None:
        """Print the menu to stdout."""
        for name in sorted(self.items.keys()):
            item = self.items[name]
            print('[{status}] {name}: {title}'.format(status=item.status, name=name, title=item.title))

        if self.parent is None:
            print('{answer} - Quit'.format(answer=CookbooksMenu.quit_answer))
        else:
            print('{answer} - Back to parent menu'.format(answer=CookbooksMenu.back_answer))

        print('{answer} - Help'.format(answer=CookbooksMenu.help_answer))

    def calculate_status(self) -> Tuple[int, int]:
        """Calculate the status of a menu, checking the status of all it's tasks recursively.

        Returns:
            tuple: (int, int) with the number of completed and total items.

        """
        completed = 0
        total = 0
        for item in self.items.values():
            if isinstance(item, CookbooksMenu):
                sub_completed, sub_total = item.calculate_status()
                completed += sub_completed
                total += sub_total
            elif isinstance(item, Cookbook):
                total += 1
                if item.status != Cookbook.not_run:
                    completed += 1
            else:  # pragma: no cover | This should never happen
                raise CookbookError('Unknown item of type {type}'.format(type=type(item)))

        return completed, total

    def get_tree(self) -> str:
        """Return the tree representation of the menu as string.

        Returns:
            str: the tree representation of all the collected items.

        """
        lines = self.get_menu_tree(0, [])
        if not lines:
            return ''

        return '{title}\n{lines}\n'.format(title=Cookbooks.cookbooks_module_prefix, lines='\n'.join(lines))

    def get_menu_tree(self, level: int, cont_levels: List[bool]) -> List[str]:
        """Calculate the tree lines for a given menu.

        Arguments:
            level (int): how many levels the item is nested in the tree.
            cont_levels (list): a list of size levels with booleans that indicate for each level if the
                continuation prefix (True) or an empty prefix (False) should be used.

        Returns:
            list: the list of lines that represent the tree.

        """
        lines = []
        for i, key in enumerate(sorted(self.items.keys())):
            is_final = (i == len(self.items) - 1)
            name = self.items[key].path
            prefix = self._get_line_prefix(level, cont_levels, is_final)

            if self.spicerack.verbose:
                line = '{prefix}{name}: {title}'.format(prefix=prefix, name=name, title=self.items[key].title)
            else:
                line = '{prefix}{name}'.format(prefix=prefix, name=name)

            lines.append(line)

            if isinstance(self.items[key], CookbooksMenu):
                lines += self.items[key].get_menu_tree(level + 1, cont_levels + [not is_final])

        return lines

    def run_once(self) -> None:
        """Run the menu in an interactive way.

        Returns:
            spicerack.cookbook.CookbooksMenu: the current menu instance.

        """
        print('#--- {title} args={args} ---#'.format(title=self.verbose_title, args=self.args))
        self.show()

        if not sys.stdout.isatty():
            print('Not a tty, exiting.')
            raise StopIteration

        try:
            answer = input('>>> ')
        except (EOFError, KeyboardInterrupt):
            print('QUIT')
            raise StopIteration  # Ctrl+d or Ctrl+c pressed while waiting for input

        if not answer:
            return

        if answer == CookbooksMenu.help_answer:
            print(self.help_message)
            return

        if answer == CookbooksMenu.quit_answer and self.parent is None:
            raise StopIteration

        if answer == CookbooksMenu.back_answer and self.parent is not None:
            raise StopIteration

        name, *args = shlex.split(answer)
        if name not in self.items.keys():
            print('==> Invalid input <==')
            return

        item = self.items[name]
        item.args = self._get_item_args(item, args)
        item.run()

    def _get_item_args(self, item: Type[BaseCookbooksItem], args: List[str]) -> List[str]:
        """Get the arguments to pass to the given item.

        Arguments:
            item (spicerack.cookbook.BaseCookbooksItem): the item to pass the arguments to.
            args (list): the arguments passed via interactive menu to this item.

        Returns:
            list: the arguments to pass to the item.

        """
        if args:  # Override the item arguments with the ones passed interactively, if any.
            return args

        if self.args:  # Override the item arguments with the ones of the menu, if any.
            return self.args

        return item.args


def argument_parser() -> argparse.ArgumentParser:
    """Get the CLI argument parser.

    If the COOKBOOK is passed as a path, it will be converted to a Python module syntax.

    Returns:
        argparse.ArgumentParser: the argument parser instance.

    """
    parser = argparse.ArgumentParser(description='Spicerack Cookbook Runner')
    parser.add_argument(
        '-l', '--list', action='store_true',
        help=('List all available cookbooks, if -v/--verbose is set print also their description. If a COOKBOOK is '
              'also specified, it will be used as a prefix filter.'))
    parser.add_argument('-c', '--config-file', default='/etc/spicerack/config.yaml',
                        help='Path to the Spicerack configuration file to load.')
    parser.add_argument('-d', '--dry-run', action='store_true', help='Set the DRY-RUN mode, also for the cookbook.')
    parser.add_argument('-v', '--verbose', action='store_true', help='Verbose output, also for the cookbook.')
    parser.add_argument(
        'cookbook', metavar='COOKBOOK', nargs='?', type=cookbook_path_type,
        help=('Either a relative path of the Python file to execute (group/cookbook.py) or the name of the Python '
              'module to execute (group.cookbook). If the selected path/module is a directory or is not set, an '
              'interactive menu will be shown.'))
    parser.add_argument(
        'cookbook_args', metavar='COOKBOOK_ARGS', nargs=argparse.REMAINDER,
        help='Collect all the remaining arguments to be passed to the cookbook or menu to execute.')

    return parser


def cookbook_path_type(path: str) -> str:
    """Convert a COOKBOOK path to module syntax, if it's in path syntax.

    Arguments:
        path (str): the path to be converted.

    Returns:
        str: the converted path in module syntax.

    """
    cookbook_path, ext = os.path.splitext(path)
    if ext == '.py':
        path = cookbook_path.replace('/', '.')

    return path


def import_module(module_name: str) -> CookbooksModuleInterface:
    """Import a Python module.

    Arguments:
        module_name (str): the name of the module to load.

    Returns:
        types.ModuleType: the loaded module, that must respect the CookbooksModuleInterface.

    Raises:
        spicerack.cookbook.CookbookError: on failure to load the module.

    """
    try:
        return importlib.import_module(module_name)  # type: ignore
    except Exception as e:  # pylint: disable=broad-except
        raise CookbookError('Failed to import module {name}: {msg}'.format(name=module_name, msg=e)) from e


def execute_cookbook(config: Dict[str, str], args: argparse.Namespace, cookbooks: Cookbooks) -> int:
    """Execute a single cookbook with its parameters.

    Arguments:
        config (dict): the configuration dictionary.
        args (argparse.Namespace): the parsed arguments.
        cookbooks (spicerack.cookbook.Cookbooks): the collected cookbooks.

    Returns:
        int: the return code, 0 on success, non-zero on cookbook failure, 98 on cookbook exception.

    """
    if args.cookbook is not None:
        path = '.'.join((Cookbooks.cookbooks_module_prefix, args.cookbook))
    else:
        path = Cookbooks.cookbooks_module_prefix

    cookbook = cookbooks.get_item(path)
    if cookbook is None:
        logger.error('Unable to find cookbook %s', args.cookbook)
        return COOKBOOK_NOT_FOUND_RETCODE

    cookbook_path, cookbook_name = os.path.split(cookbook.path.replace('.', os.sep))
    base_path = os.path.join(config['logs_base_dir'], cookbook_path)
    log.setup_logging(base_path, cookbook_name, cookbooks.spicerack.username, dry_run=args.dry_run,
                      host=config.get('tcpircbot_host', None), port=int(config.get('tcpircbot_port', 0)))

    logger.debug('Executing cookbook %s with args: %s', args.cookbook, args.cookbook_args)
    return cookbook.run()


def main(argv: Optional[List[str]] = None) -> int:
    """Entry point, run the tool.

    Arguments:
        argv (list, optional): the list of command line arguments to parse.

    Returns:
        int: the return code, zero on success, non-zero on failure.

    """
    args = argument_parser().parse_args(argv)
    config = load_yaml_config(args.config_file)
    sys.path.append(config['cookbooks_base_dir'])
    params = config.get('instance_params', {})
    params.update({'verbose': args.verbose, 'dry_run': args.dry_run})

    try:
        spicerack = Spicerack(**params)
    except TypeError as e:
        print('Unable to instantiate Spicerack, check your configuration:', e, file=sys.stderr)
        return 1

    cookbooks = Cookbooks(config['cookbooks_base_dir'], args.cookbook_args, spicerack, path_filter=args.cookbook)

    if args.list:
        print(cookbooks.menu.get_tree(), end='')
        return 0

    return execute_cookbook(config, args, cookbooks)



"""MediaWiki module tests."""
import json

from unittest import mock

import pytest
import requests

from spicerack.mediawiki import MediaWiki, MediaWikiCheckError, MediaWikiError
from spicerack.remote import RemoteExecutionError
from spicerack.tests import require_caplog, require_requests_mock


class TestMediaWiki:
    """MediaWiki class tests."""

    def setup_method(self):
        """Initialize the test environment for MediaWiki."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_confctl = mock.MagicMock()
        self.mocked_remote = mock.MagicMock()
        self.mocked_remote.query.return_value.hosts = ['host1']
        self.username = 'user1'
        self.siteinfo_url = 'http://api.svc.eqiad.wmnet/w/api.php'
        self.siteinfo_rw = {
            'batchcomplete': True,
            'query': {
                'general': {
                    'readonly': False,
                    'wmf-config': {'wmfEtcdLastModifiedIndex': 123456, 'wmfMasterDatacenter': 'eqiad'},
                },
            },
        }
        self.siteinfo_ro = {
            'batchcomplete': True,
            'query': {
                'general': {
                    'readonly': True,
                    'readonlyreason': 'MediaWiki is in read-only mode for maintenance.',
                    'wmf-config': {'wmfEtcdLastModifiedIndex': 123456, 'wmfMasterDatacenter': 'eqiad'},
                },
            },
        }

        self.mediawiki = MediaWiki(self.mocked_confctl, self.mocked_remote, self.username, dry_run=False)
        self.mediawiki_dry_run = MediaWiki(self.mocked_confctl, self.mocked_remote, self.username)

    @require_requests_mock
    def test_check_config_line(self, requests_mock):
        """It should verify the config published at noc.wikimedia.org."""
        requests_mock.get('http://host1/conf/file1.php.txt', text='data')
        assert self.mediawiki.check_config_line('file1', 'data') is True

    @require_caplog
    def test_get_siteinfo(self, requests_mock):
        """It should get the siteinfo API from a canary host."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        assert MediaWiki.get_siteinfo('eqiad') == self.siteinfo_ro

    @require_requests_mock
    def test_check_siteinfo_ok(self, requests_mock):
        """It should check that a specific key in siteinfo API matches a value and not raise exception."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_rw))
        MediaWiki.check_siteinfo('eqiad', {('query', 'general', 'wmf-config', 'wmfEtcdLastModifiedIndex'): 123456})

    @require_requests_mock
    def test_check_siteinfo_multi_ok(self, requests_mock):
        """It should check that multiple keys in siteinfo API matches multiple values and not raise exception."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_rw))
        MediaWiki.check_siteinfo('eqiad', {
            ('query', 'general', 'wmf-config', 'wmfEtcdLastModifiedIndex'): 123456,
            ('batchcomplete',): True,
            ('query', 'general', 'readonly'): False})

    @require_requests_mock
    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_check_siteinfo_raise_check(self, mocked_sleep, requests_mock):
        """It should retry if it doesn't match and raise MediaWikiCheckError after all retries have failed."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        with pytest.raises(MediaWikiCheckError, match=r"Expected 'invalid', got 'True' for path: \('batchcomplete',\)"):
            MediaWiki.check_siteinfo('eqiad', {('batchcomplete',): 'invalid'})

        assert mocked_sleep.called

    @require_requests_mock
    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_check_siteinfo_raise_error(self, mocked_sleep, requests_mock):
        """It should retry if it fails and raise MediaWikiError after all retries have failed."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        with pytest.raises(MediaWikiError, match='Failed to traverse siteinfo for key invalid'):
            MediaWiki.check_siteinfo('eqiad', {('invalid',): 'invalid'})

        assert mocked_sleep.called

    @require_caplog
    @require_requests_mock
    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_check_siteinfo_key_error(self, mocked_sleep, requests_mock):
        """It should raise MediaWikiError if the key is not present in siteinfo."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        with pytest.raises(MediaWikiError, match='Failed to traverse siteinfo for key invalid'):
            MediaWiki.check_siteinfo('eqiad', {('invalid',): 'invalid'})

        assert mocked_sleep.called

    @require_caplog
    @require_requests_mock
    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_check_siteinfo_timeout(self, mocked_sleep, requests_mock):
        """It should raise MediaWikiError if it fails to get siteinfo."""
        requests_mock.get(self.siteinfo_url, exc=requests.exceptions.ConnectTimeout)
        with pytest.raises(MediaWikiError, match='Failed to get siteinfo'):
            MediaWiki.check_siteinfo('eqiad', {('batchcomplete',): 'invalid'})

        assert mocked_sleep.called

    def test_scap_sync_config_file(self):
        """It should run scap sync file on the deployment host."""
        self.mediawiki.scap_sync_config_file('file1', 'deployed file1')
        self.mocked_remote.query.assert_called_once_with('C:Deployment::Rsync and R:Class%cron_ensure = absent')
        assert 'scap sync-file' in self.mocked_remote.query.return_value.run_sync.call_args[0][0]

    @require_requests_mock
    def test_set_readonly(self, requests_mock):
        """It should set the readonly message in Conftool and verify it in siteinfo."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        message = self.siteinfo_ro['query']['general']['readonlyreason']
        self.mediawiki.set_readonly('eqiad', message)
        self.mocked_confctl.set_and_verify.assert_called_once_with('val', message, name='ReadOnly', scope='eqiad')

    @require_requests_mock
    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_set_readonly_dry_run(self, mocked_sleep, requests_mock):
        """It should not set the readonly message in Conftool and the siteinfo check should not raise."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        self.mediawiki_dry_run.set_readonly('eqiad', 'invalid')
        assert mocked_sleep.called

    @require_requests_mock
    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_set_readonly_raise(self, mocked_sleep, requests_mock):
        """It should raise MediaWikiCheckError if unable to check the value has changed and not in dry_run."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        with pytest.raises(MediaWikiCheckError, match="Expected 'invalid', got"):
            self.mediawiki.set_readonly('eqiad', 'invalid')

        assert mocked_sleep.called

    @require_requests_mock
    def test_set_readwrite(self, requests_mock):
        """It should set the readonly variale in Conftool to False and verify it in siteinfo."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_rw))
        self.mediawiki.set_readwrite('eqiad')
        self.mocked_confctl.set_and_verify.assert_called_once_with('val', False, name='ReadOnly', scope='eqiad')

    @require_requests_mock
    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_set_readwrite_dry_run(self, mocked_sleep, requests_mock):
        """It should not set the readonly variable in Conftool and the siteinfo check should not raise."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        self.mediawiki_dry_run.set_readwrite('eqiad')
        assert mocked_sleep.called

    @require_requests_mock
    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_set_readwrite_raise(self, mocked_sleep, requests_mock):
        """It should raise MediaWikiCheckError if unable to check the value has changed and not in dry_run."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        with pytest.raises(MediaWikiCheckError, match="Expected 'False', got 'True'"):
            self.mediawiki.set_readwrite('eqiad')

        assert mocked_sleep.called

    @require_requests_mock
    def test_set_master_datacenter(self, requests_mock):
        """It should set the master datacenter in Conftool."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        requests_mock.get(self.siteinfo_url.replace('eqiad', 'codfw'), text=json.dumps(self.siteinfo_ro))
        self.mediawiki.set_master_datacenter('eqiad')
        self.mocked_confctl.set_and_verify.assert_called_once_with(
            'val', 'eqiad', name='WMFMasterDatacenter', scope='common')

    @require_requests_mock
    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_set_master_datacenter_dry_run(self, mocked_sleep, requests_mock):
        """It should not set the master datacenter in Conftool and not raise as the siteinfo check fails."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        requests_mock.get(self.siteinfo_url.replace('eqiad', 'codfw'), text=json.dumps(self.siteinfo_ro))
        self.mediawiki_dry_run.set_master_datacenter('codfw')
        assert mocked_sleep.called

    @require_requests_mock
    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_set_master_datacenter_raise(self, mocked_sleep, requests_mock):
        """It should raise MediaWikiCheckError if unable to check the value has changed and not in dry_run."""
        requests_mock.get(self.siteinfo_url, text=json.dumps(self.siteinfo_ro))
        requests_mock.get(self.siteinfo_url.replace('eqiad', 'codfw'), text=json.dumps(self.siteinfo_ro))
        with pytest.raises(MediaWikiCheckError, match="Expected 'codfw', got 'eqiad'"):
            self.mediawiki.set_master_datacenter('codfw')

        assert mocked_sleep.called

    def test_check_cronjobs_enabled(self):
        """It should ensure that the cronjobs are present and not commented out."""
        self.mediawiki.check_cronjobs_enabled('dc1')
        self.mocked_remote.query.assert_called_once_with('A:mw-maintenance and A:dc1')
        assert 'crontab -u www-data -l' in self.mocked_remote.query.return_value.run_sync.call_args[0][0]

    def test_stop_cronjobs(self):
        """It should ensure that the cronjobs are present and not commented out."""
        self.mediawiki.stop_cronjobs('dc1')
        self.mocked_remote.query.assert_called_with('A:mw-maintenance and A:dc1')
        assert 'crontab -u www-data -r' in self.mocked_remote.query.return_value.run_async.call_args[0][0].command

    def test_stop_cronjobs_stray_procs(self):
        """It should not fail is there are leftover php stray processes."""
        self.mocked_remote.query.return_value.run_sync.side_effect = [0, RemoteExecutionError(10, 'failed')]
        self.mediawiki.stop_cronjobs('dc1')



"""IPMI module.

Todo:
    replace with pyghmi.

"""
import logging

from datetime import timedelta
from subprocess import CalledProcessError, PIPE, run
from typing import List

from spicerack.decorators import retry
from spicerack.exceptions import SpicerackCheckError, SpicerackError


IPMI_PASSWORD_MAX_LEN = 20
IPMI_PASSWORD_MIN_LEN = 16
IPMI_SAFE_BOOT_PARAMS = ('0000000000', '8000020000')  # No or unimportant overrides.
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class IpmiError(SpicerackError):
    """Custom exception class for errors of the Ipmi class."""


class IpmiCheckError(SpicerackCheckError):
    """Custom exception class for check errors of the Ipmi class."""


class Ipmi:
    """Class to manage remote IPMI via ipmitool."""

    def __init__(self, password: str, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            password (str): the password to use to connect via IPMI.
            dry_run (bool, optional): whether this is a DRY-RUN.

        """
        self.env = {'IPMITOOL_PASSWORD': password}
        self._dry_run = dry_run

    def command(  # pylint: disable=no-self-use
        self,
        mgmt_hostname: str,
        command_parts: List[str],
        is_safe: bool = False
    ) -> str:
        """Run an ipmitool command for a remote management console hostname.

        Arguments:
            mgmt_hostname (str): the FQDN of the management interface of the host to target.
            command_parts (list): a list of :py:class:`str` with the IPMI command components to execute.
            is_safe (bool, optional): if this is a safe command to run also in DRY RUN mode.

        Returns:
            str: the output of the ipmitool command.

        Raises:
            spicerack.ipmi.IpmiError: on failure.

        """
        command = ['ipmitool', '-I', 'lanplus', '-H', mgmt_hostname, '-U', 'root', '-E'] + command_parts
        logger.info('Running IPMI command: %s', ' '.join(command))

        if self._dry_run and not is_safe:
            return ''

        try:
            output = run(command, env=self.env.copy(), stdout=PIPE, check=True).stdout.decode()
        except CalledProcessError as e:
            raise IpmiError('Remote IPMI for {mgmt} failed (exit={code}): {output}'.format(
                mgmt=mgmt_hostname, code=e.returncode, output=e.output)) from e

        logger.debug(output)

        return output

    def check_connection(self, mgmt_hostname: str) -> None:
        """Ensure that remote IPMI is working for the management console hostname.

        Arguments:
            mgmt_hostname (str): the FQDN of the management interface of the host to target.

        Raises:
            spicerack.ipmi.IpmiError: if unable to connect or execute a test command.

        """
        status = self.command(mgmt_hostname, ['chassis', 'power', 'status'], is_safe=True)
        if not status.startswith('Chassis Power is'):
            raise IpmiError('Unexpected chassis status: {status}'.format(status=status))

    def check_bootparams(self, mgmt_hostname: str) -> None:
        """Check if the BIOS boot parameters are back to normal values.

        Arguments:
            mgmt_hostname (str): the FQDN of the management interface of the host to target.

        Raises:
            spicerack.ipmi.IpmiCheckError: if the BIOS boot parameters are incorrect.

        """
        param = self._get_boot_parameter(mgmt_hostname, 'Boot parameter data')
        if param not in IPMI_SAFE_BOOT_PARAMS:
            raise IpmiCheckError('Expected BIOS boot params in {accepted} got: {param}'.format(
                accepted=IPMI_SAFE_BOOT_PARAMS, param=param))

    @retry(tries=3, delay=timedelta(seconds=20), backoff_mode='linear', exceptions=(IpmiCheckError,))
    def force_pxe(self, mgmt_hostname: str) -> None:
        """Force PXE for the next boot and verify that the setting was applied.

        Arguments:
            mgmt_hostname (str): the FQDN of the management interface of the host to target.

        Raises:
            spicerack.ipmi.IpmiCheckError: if unable to verify the PXE mode within the retries.

        """
        self.command(mgmt_hostname, ['chassis', 'bootdev', 'pxe'])
        boot_device = self._get_boot_parameter(mgmt_hostname, 'Boot Device Selector')
        if boot_device != 'Force PXE':
            raise IpmiCheckError('Unable to verify that Force PXE is set. The host might reboot in the current OS')

    def _get_boot_parameter(self, mgmt_hostname: str, param_label: str) -> str:
        """Get a specific boot parameter of the host.

        Arguments:
            mgmt_hostname (str): the FQDN of the management interface of the host to target.
            param_label (str): the label of the boot parameter to lookout for.

        Raises:
            spicerack.ipmi.IpmiError: if unable to find the given label or to extract its value.

        Returns:
            str: the value of the parameter.

        """
        bootparams = self.command(mgmt_hostname, ['chassis', 'bootparam', 'get', '5'], is_safe=True)
        for line in bootparams.splitlines():
            if param_label in line:
                try:
                    value = line.split(':')[1].strip(' \n')
                    break
                except IndexError:
                    raise IpmiError("Unable to extract value for parameter '{label}' from line: {line}".format(
                        label=param_label, line=line))
        else:
            raise IpmiError("Unable to find the boot parameter '{label}' in: {output}".format(
                label=param_label, output=bootparams))
        return value

    @staticmethod
    def _get_password_store_size(password: str) -> int:
        """Parse the password to determine the correct storage size.

        Ipmitool stores passwords in either 16 or 20 byte strings depending
        on the password length.

        Arguments:
            password(str): the password string to parse

        Raises:
            spicerack.ipmi.IpmiError: if unable password is too big or too small

        Returns:
            int: A number representing the storage size

        """
        if len(password) > IPMI_PASSWORD_MAX_LEN:
            raise IpmiError('New passwords is greater then the {max_len} byte limit'.format(
                max_len=IPMI_PASSWORD_MAX_LEN))
        if len(password) > IPMI_PASSWORD_MIN_LEN:
            return IPMI_PASSWORD_MAX_LEN
        if len(password) == IPMI_PASSWORD_MIN_LEN:
            return IPMI_PASSWORD_MIN_LEN
        raise IpmiError('New passwords must be {min_len} bytes minimum'.format(
            min_len=IPMI_PASSWORD_MIN_LEN))

    def reset_password(self, mgmt_hostname: str, username: str, password: str) -> None:
        """Reset the given usernames password to the one provided.

        Arguments:
            mgmt_hostname (str): the FQDN of the management interface of the host to target.
            username (str): The username who's password will be reset must not be empty
            password (str): The new password must have length between {min_len} and {max_len} bytes

        Raises:
            spicerack.ipmi.IpmiError: if unable reset password or arguments invalid

        """.format(min_len=IPMI_PASSWORD_MIN_LEN, max_len=IPMI_PASSWORD_MAX_LEN)
        if not username:
            raise IpmiError('Username can not be an empty string')

        password_store_size = str(Ipmi._get_password_store_size(password))

        try:
            user_id = self._get_user_id(mgmt_hostname, username)
        except IpmiError:
            # some systems (HP?) use channel 2
            logger.info('unable to find user in channel 1 testing channel 2')
            user_id = self._get_user_id(mgmt_hostname, username, 2)

        success = 'Set User Password command successful (user {user_id})\n'.format(
            user_id=user_id)
        result = self.command(
            mgmt_hostname,
            ['user', 'set', 'password', user_id, password, password_store_size])

        if self._dry_run:
            return

        if result != success:
            raise IpmiError('Password reset failed for username: {username}'.format(
                username=username))

        if username == 'root':
            current_password = self.env['IPMITOOL_PASSWORD']
            self.env['IPMITOOL_PASSWORD'] = password
            try:
                self.check_connection(mgmt_hostname)
            except IpmiError as error:
                self.env['IPMITOOL_PASSWORD'] = current_password
                raise IpmiError('Password reset failed for username: root') from error

    def _get_user_id(self, mgmt_hostname: str, username: str, channel: int = 1) -> str:
        """Get the user ID associated with a given username.

        Arguments:
            mgmt_hostname (str): the FQDN of the management interface of the host to target.
            username (str): The username to search for
            channel (int): The channel number for the user list; Default: 1

        Raises:
            spicerack.ipmi.IpmiError: if unable to find the given username.

        Returns:
            str: the user ID associated with the username

        """
        userlist = self.command(mgmt_hostname, ['user', 'list', str(channel)], is_safe=True)
        for line in userlist.splitlines():
            words = line.split()
            if words[0] == 'ID':
                continue
            if words[1] == username:
                return words[0]
        raise IpmiError("Unable to find ID for username: {username}".format(username=username))



"""Interactive module tests."""
import configparser
import re

from logging import DEBUG

import pytest

from spicerack.config import load_ini_config, load_yaml_config
from spicerack.exceptions import SpicerackError
from spicerack.tests import get_fixture_path, require_caplog


def test_load_yaml_config_empty():
    """Loading an empty config should return an empty dictionary."""
    config_dict = load_yaml_config(get_fixture_path('config', 'empty.yaml'))
    assert {} == config_dict


@pytest.mark.parametrize('name, message', (
    ('invalid.yaml', 'ParserError'),
    ('non-existent.yaml', 'FileNotFoundError'),
))
def test_load_yaml_config_raise(name, message):
    """Loading an invalid config should raise Exception."""
    with pytest.raises(SpicerackError, match=message):
        load_yaml_config(get_fixture_path('config', name))


@require_caplog
@pytest.mark.parametrize('name', ('invalid.yaml', 'non-existent.yaml'))
def test_load_yaml_config_no_raise(caplog, name):
    """Loading an invalid config with raises=False should return an empty dictionary."""
    with caplog.at_level(DEBUG):
        config_dict = load_yaml_config(get_fixture_path('config', name), raises=False)

    assert {} == config_dict
    assert re.search(r'DEBUG.*Could not load config file', caplog.text) is not None


def test_load_yaml_config_valid():
    """Loading a valid config should return its content."""
    config_dir = get_fixture_path('config', 'valid.yaml')
    config_dict = load_yaml_config(config_dir)
    assert 'key' in config_dict


def test_load_ini_config():
    """Loading a INI config should return a configparser.ConfigParser object."""
    config = load_ini_config(get_fixture_path('config', 'config.ini'))
    assert isinstance(config, configparser.ConfigParser)
    assert config.defaults()['key'] == 'value'



"""Dnsdisc module tests."""
from datetime import timedelta
from unittest import mock

import pytest

from spicerack.decorators import get_backoff_sleep, retry
from spicerack.exceptions import SpicerackError


class DryRunRetry:
    """Test class to simulate any Spicerack class with a ``_dry_run`` property."""

    def __init__(self, dry_run=True):
        """Initialize the ``_dry_run`` property."""
        self._dry_run = dry_run

    @retry
    def fail(self):
        """A method that fail always to trigger the retry logic."""
        raise SpicerackError('self._dry_run={d}'.format(d=self._dry_run))


def _generate_mocked_function(calls):
    func = mock.Mock()
    func.side_effect = calls
    func.__name__ = 'mocked'
    return func


@pytest.mark.parametrize('calls, sleep_calls', (
    ([True], []),
    ([SpicerackError('error1'), True], [3.0]),
    ([SpicerackError('error1'), SpicerackError('error2'), True], [3.0, 9.0]),
))
@mock.patch('spicerack.decorators.time.sleep', return_value=None)
def test_retry_pass_no_args(mocked_sleep, calls, sleep_calls):
    """Using @retry with no arguments should use the default values."""
    func = _generate_mocked_function(calls)
    ret = retry(func)()
    assert ret
    func.assert_has_calls([mock.call()] * len(calls))
    mocked_sleep.assert_has_calls([mock.call(i) for i in sleep_calls])


@pytest.mark.parametrize('dry_run', (True, False))
@mock.patch('spicerack.decorators.time.sleep', return_value=None)
def test_retry_pass_no_args_dry_run(mocked_sleep, dry_run):
    """Using @retry with no arguments should use the default values but set tries to 1 if in DRY-RUN."""
    obj = DryRunRetry(dry_run=dry_run)
    with pytest.raises(SpicerackError):
        obj.fail()

    sleep_call_count = 2
    if dry_run:
        sleep_call_count = 0
    assert mocked_sleep.call_count == sleep_call_count


@pytest.mark.parametrize('exc, calls, sleep_calls', (
    (SpicerackError, [SpicerackError('error')] * 3, [3.0, 9.0]),
    (Exception, [Exception('error')], []),
))
@mock.patch('spicerack.decorators.time.sleep', return_value=None)
def test_retry_fail_no_args(mocked_sleep, exc, calls, sleep_calls):
    """Using @retry with no arguments should raise the exception raised by the decorated function if not cathced."""
    func = _generate_mocked_function(calls)
    with pytest.raises(exc, match='error'):
        retry(func)()

    func.assert_has_calls([mock.call()] * len(calls))
    mocked_sleep.assert_has_calls([mock.call(i) for i in sleep_calls])


@pytest.mark.parametrize('calls, sleep_calls, kwargs', (
    ([True], [], {'delay': timedelta(seconds=11), 'tries': 1}),
    ([Exception('error1'), True], [5.55],
     {'delay': timedelta(seconds=5, milliseconds=550), 'tries': 2, 'exceptions': Exception}),
    ([SpicerackError('error1'), True], [8.88],
     {'backoff_mode': 'exponential', 'delay': timedelta(milliseconds=8880), 'tries': 2}),
    ([SpicerackError('error1'), True], [0.90],
     {'backoff_mode': 'power', 'delay': timedelta(milliseconds=900), 'tries': 2}),
))
@mock.patch('spicerack.decorators.time.sleep', return_value=None)
def test_retry_pass_args(mocked_sleep, calls, sleep_calls, kwargs):
    """Using @retry with arguments should use the soecified values."""
    func = _generate_mocked_function(calls)
    ret = retry(**kwargs)(func)()

    assert ret
    func.assert_has_calls([mock.call()] * len(calls))
    mocked_sleep.assert_has_calls([mock.call(i) for i in sleep_calls])


@pytest.mark.parametrize('exc, kwargs', (
    (SpicerackError, {}),
    (RuntimeError, {'exceptions': (KeyError, ValueError)}),
))
@mock.patch('spicerack.decorators.time.sleep', return_value=None)
def test_retry_fail_args(mocked_sleep, exc, kwargs):
    """Using @retry with arguments should raise the exception raised by the decorated function if not cathced."""
    func = _generate_mocked_function([exc('error')])
    kwargs['tries'] = 1
    with pytest.raises(exc, match='error'):
        retry(**kwargs)(func)()

    func.assert_called_once_with()
    assert not mocked_sleep.called


@pytest.mark.parametrize('kwargs, message', (
    ({'tries': 0}, 'Tries must be a positive integer, got 0'),
    ({'backoff_mode': 'invalid'}, 'Invalid backoff_mode: invalid'),
    ({'backoff_mode': 'exponential', 'delay': timedelta(milliseconds=500)},
     'Delay must be greater than 1 if backoff_mode is exponential'),
))
def test_retry_invalid(kwargs, message):
    """Using @retry with invalid arguments should raise ValueError."""
    with pytest.raises(ValueError, match=message):
        retry(**kwargs)(lambda: True)()  # pylint: disable=not-callable


@pytest.mark.parametrize('mode, base, values', (
    ('constant', 0, (0,) * 5),
    ('constant', 0.5, (0.5,) * 5),
    ('constant', 3, (3,) * 5),
    ('linear', 0, (0,) * 5),
    ('linear', 0.5, (0.5, 1.0, 1.5, 2.0, 2.5)),
    ('linear', 3, (3, 6, 9, 12, 15)),
    ('power', 0, (0,) * 5),
    ('power', 0.5, (0.5, 1, 2, 4, 8)),
    ('power', 3, (3, 6, 12, 24, 48)),
    ('exponential', 1, (1,) * 5),
    ('exponential', 1.5, (1.5, 2.25, 3.375, 5.0625, 7.59375)),
    ('exponential', 3, (3, 9, 27, 81, 243)),
))
def test_get_backoff_sleep(mode, base, values):
    """Calling get_backoff_sleep() should return the proper backoff based on the arguments."""
    for i, val in enumerate(values, start=1):
        assert get_backoff_sleep(mode, base, i) == pytest.approx(val)


def test_get_backoff_sleep_raise():
    """Calling get_backoff_sleep() with an invalid backoff_mode should raise ValueError."""
    with pytest.raises(ValueError, match='Invalid backoff_mode: invalid'):
        get_backoff_sleep('invalid', 1, 5)



"""Actions module tests."""
import logging

from textwrap import dedent

from spicerack import actions
from spicerack.tests import require_caplog


def test_actionsdict_string_representation():
    """It should convert the instance to a nice string representation."""
    actions_dict = actions.ActionsDict()
    actions_dict['name1'].success('success1')
    actions_dict['name2'].failure('failure1')
    expected = """
    - name1 (**PASS**)
      - success1

    - name2 (**FAIL**)
      - failure1
    """
    assert dedent(expected).lstrip() == str(actions_dict)


class TestActions:
    """Test class for the Actions class."""

    def setup(self):
        """Initialize the test instance."""
        # pylint: disable=attribute-defined-outside-init
        self.actions = actions.Actions('name1')

    def _check_logs(self, logs, message, level):  # pylint: disable=no-self-use
        """Assert that a log record with the given message and level is present."""
        for record in logs.records:
            if message in record.getMessage():
                assert record.levelno == level
                break
        else:
            raise RuntimeError("{level} log record with message '{msg}' not found".format(
                level=logging.getLevelName(level), msg=message))

    def test_default_status(self):
        """It should return a successful default status with no actions."""
        assert self.actions.status == 'PASS'
        assert not self.actions.actions

    @require_caplog
    def test_success(self, caplog):
        """It should register a success action."""
        self.actions.success('success1')
        assert self.actions.status == 'PASS'
        assert not self.actions.has_warnings
        assert not self.actions.has_failures
        assert len(self.actions.actions) == 1
        self._check_logs(caplog, 'success1', logging.INFO)

    @require_caplog
    def test_warning(self, caplog):
        """It should register a warning action."""
        self.actions.warning('warning1')
        self.actions.success('success1')
        assert self.actions.status == 'WARN'
        assert self.actions.has_warnings
        assert not self.actions.has_failures
        assert len(self.actions.actions) == 2
        self._check_logs(caplog, 'warning1', logging.WARNING)

    @require_caplog
    def test_failure(self, caplog):
        """It should register a failed action."""
        self.actions.failure('failure1')
        self.actions.success('success1')
        assert self.actions.status == 'FAIL'
        assert not self.actions.has_warnings
        assert self.actions.has_failures
        assert len(self.actions.actions) == 2
        self._check_logs(caplog, 'failure1', logging.ERROR)

    def test_failure_and_warning(self):
        """With a failure and a warning it should have a failed status but also report warnings."""
        self.actions.failure('failure1')
        self.actions.warning('warning1')
        self.actions.success('success1')
        assert self.actions.status == 'FAIL'
        assert self.actions.has_warnings
        assert self.actions.has_failures
        assert len(self.actions.actions) == 3

    def test_string_representation(self):
        """It should convert the instance to a nice string representation."""
        self.actions.success('success1')
        self.actions.warning('warning1')
        self.actions.failure('failure1')
        expected = """
        name1 (**FAIL**)
          - success1
          - warning1
          - failure1
        """
        assert dedent(expected).strip() == str(self.actions)



"""Debmonitor module tests."""
import pytest
import requests

from spicerack import debmonitor
from spicerack.tests import require_requests_mock


DEBMONITOR_HOST = 'debmonitor.example.com'
HOST1_URL = 'https://{host}/hosts/host1.example.com'.format(host=DEBMONITOR_HOST)


@require_requests_mock
class TestDebmonitor:
    """Debmonitor class tests."""

    def setup_method(self):
        """Initialize the test environment for Debmonitor."""
        # pylint: disable=attribute-defined-outside-init
        self.debmonitor = debmonitor.Debmonitor(DEBMONITOR_HOST, 'cert.pem', 'key.pem', dry_run=False)
        self.debmonitor_dry_run = debmonitor.Debmonitor(DEBMONITOR_HOST, 'cert.pem', 'key.pem')

    def test_host_delete_ok(self, requests_mock):
        """It should delete the host from Debmonitor."""
        requests_mock.delete(HOST1_URL, status_code=requests.codes['no_content'])
        self.debmonitor.host_delete('host1.example.com')
        assert requests_mock.call_count == 1
        assert requests_mock.request_history[0].method == 'DELETE'
        assert requests_mock.request_history[0].url == HOST1_URL

    def test_host_delete_dry_run(self, requests_mock):
        """It should not delete the host from Debmonitor in DRY-RUN."""
        self.debmonitor_dry_run.host_delete('host1.example.com')
        assert not requests_mock.called

    def test_host_delete_not_found(self, requests_mock):
        """It should not raise if the host is already absent in Debmonitor."""
        requests_mock.delete(HOST1_URL, status_code=requests.codes['not_found'])
        self.debmonitor.host_delete('host1.example.com')
        assert requests_mock.call_count == 1

    def test_host_delete_fail(self, requests_mock):
        """It should raise DebmonitorError if failing to delete the host from Debmonitor."""
        requests_mock.delete(HOST1_URL, status_code=requests.codes['bad_request'])
        with pytest.raises(debmonitor.DebmonitorError,
                           match='Unable to remove host host1.example.com from Debmonitor'):
            self.debmonitor.host_delete('host1.example.com')

        assert requests_mock.call_count == 1



"""Config module."""
import logging

from configparser import ConfigParser
from typing import Dict

import yaml

from spicerack.exceptions import SpicerackError


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def load_yaml_config(config_file: str, raises: bool = True) -> Dict:
    """Parse a YAML config file and return it, optionally not failing on error.

    Arguments:
        config_file (str): the path of the configuration file.
        raises (bool, optional): whether to raise exception if unable to load the config.

    Returns:
        dict: the parsed config or an empty dictionary as a fallback when ``raises`` is ``False``.

    Raises:
        SpicerackError: if unable to load the configuration and ``raises`` is ``True``.

    """
    config = {}
    try:
        with open(config_file, 'r') as fh:
            config = yaml.safe_load(fh)

    except Exception as e:  # pylint: disable=broad-except
        message = "Could not load config file %s: %s"
        if raises:
            raise SpicerackError(repr(e)) from e

        logger.debug(message, config_file, e)

    if config is None:
        config = {}

    return config


def load_ini_config(config_file: str) -> ConfigParser:
    """Parse an INI config file and return it.

    Arguments:
        config_file (str): the path of the configuration file.

    Returns:
        configparser.ConfigParser: the parsed config.

    """
    config = ConfigParser()
    config.read(config_file)

    return config



"""Decorators module."""
import logging
import time

from datetime import timedelta
from functools import wraps
from typing import Any, Callable, Optional, Tuple, Type, Union

from spicerack.exceptions import SpicerackError


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def ensure_wrap(func: Callable) -> Callable:
    """Decorator to wrap other decorators to allow to call them both with and without arguments.

    Arguments:
        func: the decorated function, it must be a decorator. A decorator that accepts only one positional argument
            that is also a callable is not supported.

    """
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Callable:
        """Decorator wrapper."""
        if len(args) == 1 and not kwargs and callable(args[0]):  # Called without arguments
            return func(args[0])

        return lambda real_func: func(real_func, *args, **kwargs)  # Called with arguments

    return wrapper


# TODO: 'func=None' is a workaround for https://github.com/PyCQA/pylint/issues/259, restore it to 'func, *' once fixed
# and remove the type: ignore comments.
@ensure_wrap
def retry(
    func: Optional[Callable] = None,
    tries: int = 3,
    delay: timedelta = timedelta(seconds=3),
    backoff_mode: str = 'exponential',
    exceptions: Tuple[Type[Exception], ...] = (SpicerackError,)
) -> Callable:
    """Decorator to retry a function or method if it raises certain exceptions with customizable backoff.

    Note:
        The decorated function or method must be idempotent to avoid unwanted side effects.
        It can be called with or without arguments, in the latter case all the default values will be used.

    Note:
        When the decorated function is an instance method of a class, the decorator is able to automatically detect if
        there is a ``self._dry_run`` property in the instance or a ``self._remote_hosts._dry_run`` one and reduce the
        ``tries`` attempts to ``1`` if it's a DRY-RUN to avoid unnecessary waits.

    Arguments:
        func (function, method): the decorated function.
        tries (int, optional): the number of times to try calling the decorated function or method before giving up.
            Must be a positive integer.
        delay (datetime.timedelta, optional): the initial delay for the first retry, used also as the base for the
            backoff algorithm.
        backoff_mode (str, optional): the backoff mode to use for the delay, available values are::

            constant:    delay       => 3, 3,  3,  3,   3, ...;
            linear:      delay * N   => 3, 6,  9, 12,  15, ...; N in [1, tries]
            power:       delay * 2^N => 3, 6, 12, 24,  48, ...; N in [0, tries - 1]
            exponential: delay^N     => 3, 9, 27, 81, 243, ...; N in [1, tries], delay must be > 1.

        exceptions (type, tuple, optional): the decorated function call will be retried if it fails until it succeeds
            or `tries` attempts are reached. A retryable failure is defined as raising any of the exceptions listed.

    Returns:
        function: the decorated function.

    """
    @wraps(func)  # type: ignore
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        """Decorator."""
        if backoff_mode not in ('constant', 'linear', 'power', 'exponential'):
            raise ValueError('Invalid backoff_mode: {mode}'.format(mode=backoff_mode))

        if backoff_mode == 'exponential' and delay.total_seconds() < 1:
            raise ValueError(
                'Delay must be greater than 1 if backoff_mode is exponential, got {delay}'.format(delay=delay))

        if tries < 1:
            raise ValueError('Tries must be a positive integer, got {tries}'.format(tries=tries))

        effective_tries = _get_effective_tries(tries, args)
        attempt = 0
        while attempt < effective_tries - 1:
            attempt += 1
            try:
                # Call the decorated function or method
                return func(*args, **kwargs)  # type: ignore
            except exceptions as e:
                sleep = get_backoff_sleep(backoff_mode, delay.total_seconds(), attempt)
                logger.warning("Failed to call '%s.%s' [%d/%d, retrying in %.2fs]: %s",
                               func.__module__, func.__name__, attempt, effective_tries, sleep, e)  # type: ignore
                time.sleep(sleep)

        return func(*args, **kwargs)  # type: ignore

    return wrapper


def get_backoff_sleep(backoff_mode: str, base: Union[int, float], index: int) -> Union[int, float]:
    """Calculate the amount of sleep for this attempt.

    Arguments:
        backoff_mode (str): the backoff mode to use for the delay, see the documentation for retry().
        base (int, float): the base for the backoff algorithm.
        index (int): the index to calculate the Nth sleep time for the backoff.

    Return:
        int, float: the amount of sleep to perform for the backoff.

    """
    if backoff_mode == 'constant':
        sleep = base
    elif backoff_mode == 'linear':
        sleep = base * index
    elif backoff_mode == 'power':
        sleep = base * 2 ** (index - 1)
    elif backoff_mode == 'exponential':
        sleep = base ** index
    else:
        raise ValueError('Invalid backoff_mode: {mode}'.format(mode=backoff_mode))

    return sleep


def _get_effective_tries(tries: int, decorator_args: Tuple) -> int:
    """Try to detect if this is a DRY-RUN and reduce the number of tries to one in that case.

    Arguments:
        tries (int): the requested number of tries for the decorator.
        decorator_args (tuple): tuple of positional arguments passed to the decorated function.

    Returns:
        int: the number of tries to use given the requested tries and the detection if this is a DRY-RUN or not.

    """
    effective_tries = tries
    if not decorator_args:
        logger.debug('Decorator called without args')
        return effective_tries

    obj = decorator_args[0]
    if getattr(obj, '_dry_run', False) or getattr(getattr(obj, '_remote_hosts', False), '_dry_run', False):
        logger.warning('Reduce tries from %d to 1 in DRY-RUN mode', tries)
        effective_tries = 1

    return effective_tries



"""Ganeti Module test."""

import json

from unittest import mock

import pytest
import requests

from ClusterShell.MsgTree import MsgTreeElem
from cumin import NodeSet

from spicerack.exceptions import SpicerackError
from spicerack.ganeti import CLUSTERS_AND_ROWS, Ganeti, GanetiError, GanetiRAPI, GntInstance, RAPI_URL_FORMAT
from spicerack.remote import Remote
from spicerack.tests import get_fixture_path, require_requests_mock


@require_requests_mock
class TestGaneti:
    """Ganeti tests class."""

    def setup_method(self):
        """Setup test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.remote = mock.MagicMock(spec_set=Remote)
        self.ganeti = Ganeti(username='user', password='pass', timeout=10, remote=self.remote)  # nosec

        self.cluster = 'ganeti01.svc.eqiad.wmnet'
        self.instance = 'test.example.com'
        self.cluster_base_url = {cluster: RAPI_URL_FORMAT.format(cluster=cluster) + '/2'
                                 for cluster in CLUSTERS_AND_ROWS}
        self.base_url = self.cluster_base_url[self.cluster]
        self.instance_url = '{base_url}/instances/{instance}'.format(base_url=self.base_url, instance=self.instance)

        # load test fixtures
        with open(get_fixture_path('ganeti', 'info.json'), encoding='utf-8') as info_json:
            self.info = info_json.read()
        with open(get_fixture_path('ganeti', '404.json'), encoding='utf-8') as fourohfour_json:
            self.fourohfour = fourohfour_json.read()
        with open(get_fixture_path('ganeti', 'instance.json'), encoding='utf-8') as instance_json:
            self.instance_info = instance_json.read()
        with open(get_fixture_path('ganeti', 'bogus.json'), encoding='utf-8') as bogus_json:
            self.bogus_data = bogus_json.read()

    def _set_requests_mock_for_instance(self, requests_mock, missing_active=False):
        """Set request mock to be 404 on all other clusters."""
        for cluster in CLUSTERS_AND_ROWS:
            if cluster == self.cluster and not missing_active:
                requests_mock.get(self.instance_url, text=self.instance_info)
            else:
                requests_mock.get('{base_url}/instances/{instance}'.format(
                    base_url=self.cluster_base_url[cluster], instance=self.instance),
                    text=self.fourohfour, status_code=requests.codes['not_found'])

    @pytest.mark.parametrize('cluster', CLUSTERS_AND_ROWS.keys())
    def test_rapi_clusters_ok(self, cluster):
        """Ganeti.rapi() should return a GanetiRAPI object."""
        assert isinstance(self.ganeti.rapi(cluster), GanetiRAPI)

    def test_rapi_clusters_missing(self):
        """If a cluster that doesn't exist is passed to Ganeti.rapi(), it should except with a SpicerackError."""
        with pytest.raises(SpicerackError, match=r'Cannot find cluster bogus cluster \(expected .*'):
            self.ganeti.rapi('bogus cluster')

    def test_rapi_master_ok(self, requests_mock):
        """The master property of a RAPI should be the hostname for the master of this cluster."""
        requests_mock.get(self.base_url + '/info', text=self.info)

        master = json.loads(self.info)['master']
        rapi = self.ganeti.rapi(self.cluster)
        assert rapi.master == master

    def test_rapi_master_missing(self, requests_mock):
        """If a master is not specified by the uptsream API, the value of master on a RAPI should be None."""
        requests_mock.get(self.base_url + '/info', text=self.bogus_data)
        rapi = self.ganeti.rapi(self.cluster)
        assert rapi.master is None

    def test_rapi_instance_missing(self, requests_mock):
        """A RAPI object should raise a GanetiError if a requested host does not exist."""
        requests_mock.get(self.instance_url, text=self.fourohfour, status_code=requests.codes['not_found'])
        rapi = self.ganeti.rapi(self.cluster)
        with pytest.raises(GanetiError, match=r'Non-200 from API: 404:.*'):
            rapi.fetch_instance(self.instance)

    def test_rapi_instance_timeout(self, requests_mock):
        """A RAPI object should raise a GanetiError if a request times out."""
        rapi = self.ganeti.rapi(self.cluster)
        requests_mock.get(self.instance_url, exc=requests.exceptions.ConnectTimeout)
        with pytest.raises(GanetiError, match='Timeout performing request to RAPI'):
            rapi.fetch_instance(self.instance)

    def test_rapi_instance_invalid(self, requests_mock):
        """If no mac is present in host data, fetch_instance_mac should raise a GanetiError."""
        rapi = self.ganeti.rapi(self.cluster)
        requests_mock.get(self.instance_url, text=self.bogus_data)
        with pytest.raises(GanetiError, match=''):
            rapi.fetch_instance_mac(self.instance)

    def test_rapi_instance_valid(self, requests_mock):
        """The MAC returned by RAPI.fetch_instance_mac should match the data returned by the API."""
        rapi = self.ganeti.rapi(self.cluster)
        requests_mock.get(self.instance_url, text=self.instance_info)
        mac = json.loads(self.instance_info)['nic.macs'][0]
        assert mac == rapi.fetch_instance_mac(self.instance)

    def test_fetch_cluster_for_instance_ok(self, requests_mock):
        """We should get a cluster name that the host exists for."""
        self._set_requests_mock_for_instance(requests_mock)
        assert self.ganeti.fetch_cluster_for_instance(self.instance) == self.cluster

    def test_fetch_cluster_for_instance_missing(self, requests_mock):
        """We should get an exception if the host is not found in Ganeti."""
        self._set_requests_mock_for_instance(requests_mock, missing_active=True)
        with pytest.raises(GanetiError, match='Cannot find test.example.com in any configured cluster.'):
            self.ganeti.fetch_cluster_for_instance(self.instance)

    @pytest.mark.parametrize('cluster', ('', 'ganeti01.svc.eqiad.wmnet'))
    def test_instance_ok(self, cluster, requests_mock):
        """It should return an instance of GntInstance for a properly configured cluster."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + '/info', text=self.info)
        instance = self.ganeti.instance(self.instance, cluster=cluster)
        assert isinstance(instance, GntInstance)
        assert instance.cluster == 'ganeti01.svc.eqiad.wmnet'
        self.remote.query.assert_called_once_with('ganeti1.example.com')

    def test_instance_missing_master(self, requests_mock):
        """It should raise a GanetiError exception if unable to determin the instance's master to manage it."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + '/info', text=self.bogus_data)
        with pytest.raises(GanetiError, match='Master for cluster ganeti01.svc.eqiad.wmnet is None'):
            self.ganeti.instance(self.instance)

    @pytest.mark.parametrize('kwargs', ({}, {'timeout': 0}))
    def test_instance_shutdown(self, requests_mock, kwargs):
        """It should issue the shutdown command on the master host."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + '/info', text=self.info)
        instance = self.ganeti.instance(self.instance)
        instance.shutdown(**kwargs)
        timeout = kwargs['timeout'] if 'timeout' in kwargs else 2
        self.remote.query.return_value.run_sync.assert_called_once_with(
            'gnt-instance shutdown --timeout={timeout} test.example.com'.format(timeout=timeout))

    @pytest.mark.parametrize('kwargs', ({}, {'shutdown_timeout': 0}))
    def test_instance_remove(self, requests_mock, kwargs):
        """It should issue the remove command on the master host."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + '/info', text=self.info)
        instance = self.ganeti.instance(self.instance)
        instance.remove(**kwargs)
        timeout = kwargs['shutdown_timeout'] if 'shutdown_timeout' in kwargs else 2
        self.remote.query.return_value.run_sync.assert_called_once_with(
            'gnt-instance remove --shutdown-timeout={timeout} --force test.example.com'.format(timeout=timeout))

    def test_instance_add_ok(self, requests_mock):
        """It should issue the remove command on the master host."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + '/info', text=self.info)
        instance = self.ganeti.instance(self.instance)
        results = [(NodeSet('ganeti-master.example.com'), MsgTreeElem(b'creation logs', parent=MsgTreeElem()))]
        self.remote.query.return_value.run_sync.return_value = iter(results)

        instance.add(row='A', vcpus=2, memory=3, disk=4, link='private')

        self.remote.query.return_value.run_sync.assert_called_once_with(
            'gnt-instance add -t drbd -I hail --net 0:link=private --hypervisor-parameters=kvm:boot_order=network '
            '-o debootstrap+default --no-install -g row_A -B vcpus=2,memory=3g --disk 0:size=4g test.example.com')

    @pytest.mark.parametrize('kwargs, exc_message', (
        ({'row': 'A', 'vcpus': 1, 'memory': 1, 'disk': 1, 'link': 'invalid'},
         r"Invalid link 'invalid', expected one of: \('public', 'private', 'analytics'\)"),
        ({'row': 'invalid', 'vcpus': 1, 'memory': 1, 'disk': 1, 'link': 'private'},
         r"Invalid row 'invalid' for cluster ganeti01.svc.eqiad.wmnet, expected one of: \('A', 'B', 'C', 'D'\)"),
        ({'row': 'A', 'vcpus': -1, 'memory': 1, 'disk': 1, 'link': 'private'},
         r"Invalid value '-1' for vcpus, expected positive integer."),
        ({'row': 'A', 'vcpus': 1, 'memory': -1, 'disk': 1, 'link': 'private'},
         r"Invalid value '-1' for memory, expected positive integer."),
        ({'row': 'A', 'vcpus': 1, 'memory': 1, 'disk': -1, 'link': 'private'},
         r"Invalid value '-1' for disk, expected positive integer."),
    ))
    def test_instance_add_fail(self, requests_mock, kwargs, exc_message):
        """It should raise GanetiError on invalid parameters."""
        self._set_requests_mock_for_instance(requests_mock)
        requests_mock.get(self.base_url + '/info', text=self.info)
        instance = self.ganeti.instance(self.instance)
        with pytest.raises(GanetiError, match=exc_message):
            instance.add(**kwargs)

        assert not self.remote.query.return_value.run_sync.called
